structure Geometry =
  struct
    datatype rectangle = Rectangle of { x: int,y: int,width: int,height: int }
                                       (* upper left corner *)
    datatype circle = Circle of { x: int, y: int, radius: int }
                                  (* center *)

    fun Whirlpool (x, y) = Circle {x=x, y=y, radius=30}

    fun dist (x1, y1) (x2, y2) =
        let fun r z = real z
            fun i z = round z
        in
            i(Math.sqrt (Math.pow (r(x2 - x1), 2.0)
                       + Math.pow (r(y2 - y1), 2.0)))
        end

    fun realdist (x1, y1) (x2, y2) =
        Math.sqrt (Math.pow (x2 - x1, 2.0) + Math.pow (y2 - y1, 2.0))

    infix 3 -- ++
    infix 5 ** //

    fun (x1:real, y1:real) -- (x2, y2) = (x1 - x2, y1 - y2)
    fun (x1:real, y1:real) ++ (x2, y2) = (x1 + x2, y1 + y2)
    fun (n:real) ** (x, y) = (n * x, n * y)
    fun (x:real, y:real) // n = (x / n, y / n)

    fun roundvec (x, y) = (round x, round y)

    infix 3 --- +++
    infix 5 *** ///
    fun (x1, y1) --- (x2, y2) = (x1 - x2, y1 - y2)
    fun (x1, y1) +++ (x2, y2) = (x1 + x2, y1 + y2)
    fun (n) *** (x, y) = (n * x, n * y)
    fun (x, y) /// n = (x / n, x / n)

    fun realnorm (x, y) =
        (x, y) // Math.sqrt (x * x + y * y)

    fun intersects (Circle {x=cx, y=cy, radius})
                   (Rectangle {x=rx, y=ry, width=w, height=h}) =
            let             fun in_circle (x, y) = dist (cx, cy) (x, y) < radius
                fun in_rect (x, y) = x >= rx andalso x < rx + w
                             andalso y >= ry andalso y < ry + h
                (* rectangle corners *)
                val nw = (rx, ry)
                val ne = (rx + w - 1, ry)
                val sw = (rx, ry + h - 1)
                val se = (rx + w - 1, ry + h - 1)
                (* circle diamond points, and center *)
                val center = (cx, cy)
                val top = (cx, cy - radius + 10)
                val bot = (cx, cy + radius - 10)
                val left = (cx - radius + 10, cy)
                val right = (cx + radius - 10, cy)
            in
                (* DEBUG *)
                (*
                let fun draw (x, y) = SDL.drawpixel (screen, x, y, red)
                    fun ward (x, y) = SDL.drawpixel (screen, x, y, green)
                in
                    draw center; draw top; draw bot; draw left; draw right;
                    ward nw; ward ne; ward sw; ward se;
                    SDL.flip screen
                end;
                *)
                in_circle nw orelse in_circle ne orelse
                in_circle sw orelse in_circle se orelse
                in_rect top orelse in_rect bot orelse
                in_rect left orelse in_rect right orelse
                in_rect center
            end
  end

open Geometry
infix 3 -- ++
infix 5 ** //

(* SDL.getticks () -- how many hundredths of a second since SDL started? *)
(* event loop: a case statement *)

val screen = SDL.makescreen (700, 700)

fun requireimage s =
      let val s = "testgraphics/" ^ s
      in
          case SDL.Image.load s of
              NONE => (print ("couldn't open " ^ s ^ "\n");
                       raise Fail "end of death")
            | SOME p => p
      end

structure Font =
    FontFn (val surf = requireimage "font.png"
            val charmap =
            " ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789" ^
            "`-=[]\\;',./~!@#$%^&*()_+{}|:\"<>?" (* \" *)
            (* CHECKMARK ESC HEART LCMARK1 LCMARK2 BAR_0 BAR_1 BAR_2 BAR_3-
               BAR_4 BAR_5 BAR_6 BAR_7 BAR_8 BAR_9 BAR_10 BARSTART LRARROW
               LLARROW *)
            val width = 9
            val height = 16
            val styles = 6
            val overlap = 1
            val dims = 3)

(* foreground colors for debugging *)
val red = SDL.color (0wxff, 0w0, 0w0, 0w0)
val green = SDL.color (0w0, 0wxff, 0w0, 0w0)

(* background color *)
val blue = SDL.color (0w0, 0w0, 0w0, 0w100)

(*
(* XXX DEBUGGING *)
structure SDL =
  struct
    open SDL
    (* fun fillrect _ = () *)
    fun blitall (_, screen, x, y) = fillrect (screen, x, y, 3, 3, red)
  end
*)

fun status s = () (* Font.draw (screen, 10, 10, s) *)

val starfish = requireimage "starfish.png"
val whirlpool_ccw = requireimage "whirlpool-ccw.png"
val island = requireimage "island.png"

fun getkey () =
    case SDL.pollevent () of
        NONE => (SDL.delay 0; getkey ())
      | SOME (SDL.E_KeyDown _) => ()
      | SOME _ => (SDL.delay 0; getkey ())

fun messagebox s =
    let val alphagrey = SDL.color (0w100, 0w100, 0w100, 0w50)
        val white = SDL.color (0wxff, 0wxff, 0wxff, 0w0)
        val textwidth = Font.sizex s
        val boxwidth = textwidth + 2 * 50
        val boxheight = 100
        val (x, y) = (SDL.surface_width screen div 2 - boxwidth div 2,
                      SDL.surface_height screen div 2 - boxheight div 2)
        val (tx, ty) = (x + 50, y + 42)
        val (okx, oky) = (x + boxwidth - 62, y + boxheight - 26)
    in
        SDL.fillrect (screen, x+10, y+10, boxwidth, boxheight, alphagrey);
        SDL.fillrect (screen, x, y, boxwidth, boxheight, white);
        Font.draw (screen, tx, ty, s);
        SDL.flip screen;
        SDL.delay 500;
        Font.draw (screen, okx, oky, "[ ok ]");
        SDL.flip screen
    end

fun win () = (messagebox "^0Y^1o^2u ^3m^4a^5d^6e ^7i^8t^9!"; getkey ())
fun lose () = (messagebox "^3Oh no!  Swallowed by a whirlpool!"; getkey ())

fun draw_vector (x, y) (x', y') =
    let in
        SDL.drawpixel (screen, x, y, red);
        SDL.drawpixel (screen, x', y', green)
    end
        handle SDL.SDL _ => ()

fun draw_goal (SOME (Rectangle {x, y, width, height})) =
        let val grey = SDL.color (0w100, 0w100, 0w100, 0w0)
        in
            (* SDL.fillrect (screen, x, y, width, height, grey) *)
            SDL.blitall (island, screen, x, y - 50)
        end
  | draw_goal (_) = ()

fun draw_whirlpool (Circle {x, y, radius=rad}) =
        let val (nwx, nwy) = (x - rad, y - rad)
        in
            SDL.blitall (whirlpool_ccw, screen, nwx, nwy)
        end

(* XXX DEBUG *)
val forcevector = ref (0.0, 0.0)

datatype state = S of { t : Word32.word,    (* ticks since start *)
                        x : real, y : real, (* coords of starfish *)
                        vx: real, vy: real, (* velocity of starfish *)
                        ax: real, ay: real, (* accel of starfish *)
                        lt: bool, rt: bool, (* status of left/right keys *)
                        up: bool, dn: bool, (* status up up/down keys *)
                        goal: rectangle option, (* location of goal *)
                        whirlpools: circle list (* locations of whirlpools *)
                      }

fun render (state as S {t, x, y, vx, vy, ax, ay, lt, rt, up, dn, goal, whirlpools}) =
    let fun r2s r = StringCvt.padLeft #" " 6 (Real.fmt (StringCvt.FIX (SOME 2)) r)
        fun a2s r = StringCvt.padLeft #" " 6 (Real.fmt (StringCvt.FIX (SOME 3)) r)
        fun b2s b = if b then "*" else "_"
    in
    (SDL.clearsurface (screen, blue);
     (*
     status ("x=" ^ r2s x ^ ",y=" ^ r2s y
          ^ ",vx=" ^ r2s vx ^ ",vy=" ^ r2s vy
          ^ ",ax=" ^ a2s ax ^ ",ay=" ^ a2s ay
       (* ^ ",^5lt=" ^ b2s lt ^ ",rt=" ^ b2s rt
          ^ ",up=" ^ b2s up ^ ",dn=" ^ b2s dn *)
          ^ " ^4" ^ "fx=" ^ a2s (#1 (!forcevector))
                 ^ ",fy=" ^ a2s (#2 (!forcevector))
          ^ " ^3t=" ^ Int.toString (Word32.toInt t));
     *)
     draw_goal goal;
     app draw_whirlpool whirlpools;
     SDL.blitall (starfish, screen, Real.round x, Real.round y);
        (* XXX DEBUG *)
        let val r = SDL.surface_height starfish div 2
            val (x, y) = (x + real r, y + real r)
            val (fx, fy) = !forcevector
        in  draw_vector (round x, round y) (roundvec ((x, y) ++ (100000.0 ** (fx, fy))))
        end;
     SDL.flip screen)
    end

val init_state = S { t = 0w0, x = 320.0, y = 320.0, vx = 0.0, vy = 0.0,
                     ax = 0.0, ay = 0.0,
                     lt = false, rt = false, up = false, dn = false,
                     goal = NONE, whirlpools = [] }

val level_one = S { t = 0w0, x = 100.0, y = 600.0, vx = 0.0, vy = 0.0,
                    ax = 0.0, ay = 0.0,
                    lt = false, rt = false, up = false, dn = false,
                    goal = SOME (Rectangle { x=550, y=100, width=100, height=50 }),
                    whirlpools = [ Whirlpool (300, 500),
                                   Whirlpool (500, 300),
                                   Whirlpool (475, 475),
                                   Whirlpool (100, 100),
                                   Whirlpool (100, 450),
                                   Whirlpool (400, 100),
                                   Whirlpool (250, 300),
                                   Whirlpool (620, 620) ]
                  }

fun setLT (S { t, x, y, vx, vy, ax, ay, lt, rt, up, dn, goal, whirlpools }) V =
       S { t=t, x=x, y=y, vx=vx, vy=vy, ax=ax, ay=ay,
           lt=V, rt=rt, up=up, dn=dn, goal=goal, whirlpools=whirlpools }
fun setRT (S { t, x, y, vx, vy, ax, ay, lt, rt, up, dn, goal, whirlpools }) V =
       S { t=t, x=x, y=y, vx=vx, vy=vy, ax=ax, ay=ay,
           lt=lt, rt=V, up=up, dn=dn, goal=goal, whirlpools=whirlpools }
fun setUP (S { t, x, y, vx, vy, ax, ay, lt, rt, up, dn, goal, whirlpools }) V =
       S { t=t, x=x, y=y, vx=vx, vy=vy, ax=ax, ay=ay,
           lt=lt, rt=rt, up=V, dn=dn, goal=goal, whirlpools=whirlpools }
fun setDN (S { t, x, y, vx, vy, ax, ay, lt, rt, up, dn, goal, whirlpools }) V =
       S { t=t, x=x, y=y, vx=vx, vy=vy, ax=ax, ay=ay,
           lt=lt, rt=rt, up=up, dn=V, goal=goal, whirlpools=whirlpools }

fun whirlpool_force (x, y) (Circle {x=cx, y=cy, ...}) =
    let val radius = SDL.surface_height starfish div 2
        val (sx, sy) = (x + real radius, y + real radius)
        val (cx, cy) = (real cx, real cy)
        val r = realdist (cx, cy) (sx, sy)
        val magin = 0.45 / (r * r) (* inverse square law *)
        val dirin = realnorm ((cx, cy) -- (sx, sy))
        val magtan = 0.1 * magin (* 0.000001 * Math.ln r (* if r < 40.0 then 0.005 / r else 0.0 (* 0.5 * magin *) *) *)
        val dirtan = (~(#2 dirin), #1 dirin)
    in
        (magin ** dirin) ++ (magtan ** dirtan)
    end

fun update (state as S {t=t0, x, y, vx, vy, ax, ay, lt, rt, up, dn, goal, whirlpools}) =
    let fun r x = Real.round x
        fun inbounds_x x = r x >= 0 andalso r x < SDL.surface_width screen
        fun inbounds_y y = r y >= 0 andalso r y < SDL.surface_height screen
        fun dec z d = if Real.abs z < d then 0.0 else z - real(Real.sign z) * d
        (* val t = SDL.getticks () *)
        val dt = 17.0 (* Real.fromInt (Word32.toInt (t - t0)) *)
                (* 17 ms: determined experimentally *)
        (* val () = print ("dt: " ^ Int.toString (Word32.toInt (t-t0)) ^ "\n") *)
        val da = 0.00010 (* accelerator *)
        val maxa = 0.0020
        fun tr z = if z > maxa then maxa else if z < ~maxa then ~maxa else z
        val ax = tr (if lt andalso not rt then ax - da
                else if rt andalso not lt then ax + da
                else (* 0.0 *) dec ax da (* 0.95 * ax *))
        val ay = tr (if up andalso not dn then ay - da
                else if dn andalso not up then ay + da
                else (* 0.0 *) dec ay da (* 0.95 * ay *))
        (* acceleration due to whirlpools *)
        val (fx, fy) = foldr (fn (w, (fx, fy)) =>
                                (fx, fy) ++ whirlpool_force (x, y) w)
                             (0.0, 0.0)
                             whirlpools
        (* XXX DEBUG *)
        val () = forcevector := (fx, fy)
        (* add whirlpool force vectors, and cap accelerations *)
        val (ax, ay) = (ax, ay) ++ (fx, fy)
        val (ax, ay) = (tr ax, tr ay)
        val newx = x + vx * dt + 0.5 * ax * Math.pow (dt, 2.0)
        val newy = y + vy * dt + 0.5 * ay * Math.pow (dt, 2.0)
        val vx = vx + ax * dt
        val vy = vy + ay * dt
        (*
        (* additive friction -- weird, but kind of awesome :)
           very high speeds possible! *)
        val dv = 0.0005 (* friction *)
        val vx = dec vx dv
        val vy = dec vy dv
        *)
        (* multiplicative friction -- more realistic *)
        val vx = 0.95 * vx
        val vy = 0.95 * vy
        val (x, vx) = if inbounds_x newx then (newx, vx) else (x, ~vx)
        val (y, vy) = if inbounds_y newy then (newy, vy) else (y, ~vy)
    in
      (* XXX time is not doing anything at the moment *)
      S {t=t0, x=x, y=y, vx=vx, vy=vy, ax=ax, ay=ay,
         lt=lt, rt=rt, up=up, dn=dn, goal=goal, whirlpools=whirlpools}
    end

fun winning (state as S {x, y, goal = SOME (goal as Rectangle _),
                         t, vx, vy, ax, ay, lt, rt, up, dn, whirlpools }) =
        let val radius = SDL.surface_height starfish div 2
            val starfish = Circle {x=round x + radius, y=round y + radius,
                                   radius=radius}
        in
            intersects starfish goal
        end
  | winning _ = false

fun losing (state as S {x, y, whirlpools, ...}) =
        let val radius = SDL.surface_height starfish div 2
            val (cx, cy) = (round x + radius, round y + radius)
            fun in_whirlpool (Circle {x=wx, y=wy, ...}) =
                    dist (cx, cy) (wx, wy) <= radius
        in
            List.exists in_whirlpool whirlpools
        end

fun loop state =
    let
        (* val state as S {t, x, y, vx, vy, ax, ay, lt, rt, up, dn, goal, whirlpools} = update state *)
        val state = update state
        val () = render state
    in
    if winning state then
        (win (); loop level_one)
    else if losing state then
        (lose (); loop level_one)
    else
    case SDL.pollevent () of
        NONE => (SDL.delay 0; loop state)
      (* quit on escape or Mac quit *)
      | SOME (SDL.E_KeyDown { sym = SDL.SDLK_ESCAPE }) => ()
      | SOME SDL.E_Quit => ()
      (* win on w! *)
      | SOME (SDL.E_KeyDown { sym = SDL.SDLK_w }) => win ()
      (* change dx and dy on arrow keys *)
      | SOME (SDL.E_KeyDown { sym = SDL.SDLK_LEFT })  => loop (setLT state true)
      | SOME (SDL.E_KeyDown { sym = SDL.SDLK_RIGHT }) => loop (setRT state true)
      | SOME (SDL.E_KeyDown { sym = SDL.SDLK_UP })    => loop (setUP state true)
      | SOME (SDL.E_KeyDown { sym = SDL.SDLK_DOWN })  => loop (setDN state true)
      | SOME (SDL.E_KeyUp { sym = SDL.SDLK_LEFT })  => loop (setLT state false)
      | SOME (SDL.E_KeyUp { sym = SDL.SDLK_RIGHT }) => loop (setRT state false)
      | SOME (SDL.E_KeyUp { sym = SDL.SDLK_UP })    => loop (setUP state false)
      | SOME (SDL.E_KeyUp { sym = SDL.SDLK_DOWN })  => loop (setDN state false)
      (* levels *)
      | SOME (SDL.E_KeyDown { sym = SDL.SDLK_0 }) => loop init_state
      | SOME (SDL.E_KeyDown { sym = SDL.SDLK_1 }) => loop level_one
      | SOME _ => loop state
    end

val () = (messagebox "^3Welcome to ^5Bouncecrab, Starfish Island!^<  Can you get home?";
          getkey ();
          (* loop init_state *)
          loop level_one)
